/*
*	name: cpdir
*	os: plan9 v4
*	version: 1.4
*	usage: cpdir [-mugvR] [-l file] [-x xfile] srcdir dstdir [path ...]
*		srcdir: source directory
*		dstdir: destination directory
*		path: path to copy. file or directory.
*	options:
*		-m:	merge (this option is used if dstdir is already exist)
*		-u: copy owner info.
*		-g: copy groupe info.
*		-v: verbose
*		-R: remove redundant files in destination
*		-l file: path list file
*		-x xfile: path list file that should not be copied.
*	path list file is a file that contains path per line.
*
*	# example1: synchronize /n/9fs to /n/kfs
*		disk/kfscmd allow
*		cp -x /n/9fs/adm/users /n/kfs/adm/users
*		disk/kfscmd user
*		cpdir -mugR /n/9fs /n/kfs
*		disk/kfscmd disallow
*
*	# example2: make a new backup of /n/9fs/usr/arisawa
*	   to /n/kfs/usr/arisawa/V3
*		cpdir -g /n/9fs/usr/arisawa /n/kfs/usr/arisawa/V3
*
*	Bug:
*	- The source code is too redundant.
*
*	Note:
*		Not enough tested for -m option.
*
*	auther: Kenji Arisawa (Kenar)
*	E-mail: arisawa@aichi-u.ac.jp
*/

#include <u.h>
#include <libc.h>
#include <bio.h>
#include <ctype.h>
#define MAXPATH 4096
#define debug if(Dflag) print

char *mkpath2(char *path1, char *path2);
char *mkpath3(char *path1, char *path2, char *path3);
char *strtrim(char *s);
int growto(Dir **dirbuf, int ndirbuf, long n);
int compar(Dir *a, Dir *b);
void cpdir(char *s);
int check(char *path);
int dirinfo(Dir **dirbuf, char *path);
int dirfinfo(Dir **dirbuf, int fd);
int mkdir(char *dstpath);
void cpfile(char *path);
int	copyfile(int sfd, int dfd);
int	copydirinfo(Dir *sd, Dir *dd);
int rmdir(char *path);
int rmfile(char *path);

char *src, *dst;
int vflag; // verbose
int Rflag; // remove redundant files
int uflag; // user flag
int gflag; // groupe flag
int mflag; // merge flag
int Dflag; // debug

char *usage="usage: cpdir [-mugvR] [-l file] [-x xfile] srcdir dstdir [path ....]";

typedef struct List List;
struct List{
	char	*data;
	List	*next;
};

List *xlist;

void
appendList(List **j, char *d)
{
	List *p, *q;
	q = nil;
	p = *j;
	while(p){
		q = p;
		p = p->next;
	}
	// then p is nil
	p = malloc(sizeof(List));
	if(q)
		q ->next = p;
	else
		*j = p;
	p->data = strdup(d);
	p->next = nil;
}

#ifdef NOTE
*	shell style path matching function
*	I modified to simplify original one
*	and made some modification of the matching rule.
*	now, '/' is a simplly regular charactor except
*	"/*/" is match to "/".
*	ref: /sys/src/cmd/rc/glob.c
*	-Kenar-
*
 * 	Does the string s match the pattern p
 * 	* matches any sequence of characters
 * 	? matches any single character
 * 	[...] matches the enclosed list of characters
 *	~ in the beginning of [...] means compliment
 *	- in [...] means range in ascii order
 *
 *	return value
 *	1: matched
 *	0: not matched
#endif
int
match(char *s, char *p)
{
	char *s0;
	int compl, hit, lo, hi, t, c;
	s0 = s;
	for(;*p; s++,p++){
		switch(*p){
		case '*':
			++p;
			if(*p == '/' && s > s0)
				if(match(s-1, p)) return 1;
			for(;;){
				if(match(s, p)) return 1;
				if(!*s) break;
				s++;
			}
			return 0;
		case '?':
			if(*s=='\0') return 0;
			break;
		case '[':
			if(*s=='\0') return 0;
			c=*s;
			p++;
			compl=*p=='~';
			if(compl) p++;
			hit=0;
			while(*p!=']'){
				if(*p=='\0') return 0;		/* syntax error */
				lo=*p;
				p++;
				if(*p!='-') hi=lo;
				else{
					p++;
					if(*p=='\0') return 0;	/* syntax error */
					hi=*p;
					p++;
					if(hi<lo){ t=lo; lo=hi; hi=t; }
				}
				if(lo<=c && c<=hi) hit=1;
			}
			if(compl) hit=!hit;
			if(!hit) return 0;
			break;
		default:
			if(*p != *s) return 0;
		}
	}
	return *s=='\0';
}



void *
emalloc(ulong n)
{
	void *p;

	if((p = malloc(n)) == 0)
		sysfatal("out of memory");
	return p;
}

char *
mkpath2(char *path1, char *path2)
{
	char *p;
	p = emalloc(strlen(path1)+strlen(path2)+5);
	sprint(p, "%s/%s", path1, path2);
	p = cleanname(p);
	return p;
}

char *
mkpath3(char *path1, char *path2, char *path3)
{
	char *p;
	p = emalloc(strlen(path1)+strlen(path2)+strlen(path3)+5);
	sprint(p, "%s/%s/%s", path1, path2, path3);
	p = cleanname(p);
	return p;
}

char*
strtrim(char *s)
{
	char *t;
	while(*s && isspace(*s))
		s++;
	t = s + strlen(s) - 1;
	if(s < t){
		while(*t && isspace(*t)) t--;
		t++;
		*t = 0;
	}
	return s;
}

int
xmatch(char *path)
{
	List *p;
	p = xlist;
	while(p){
		if(match(path, p->data))
			return 1;
		p = p->next;
	}
	return 0;
}

void
main(int argc, char *argv[])
{
	char *file = 0;
	char *xfile = 0;
	char *path;
	USED(argc);
	ARGBEGIN{
	case 'l':	file = ARGF(); break;
	case 'v':	vflag = 1; break;
	case 'R':	Rflag = 1; break;
	case 'u':	uflag = 1; break;
	case 'g':	gflag = 1; break;
	case 'm':	mflag = 1; break;
	case 'x':	xfile = ARGF(); break;
	case 'D':	Dflag = 1; break;
	default: sysfatal(usage);
	}ARGEND
	if(*argv)
		src = cleanname(*argv++);
	if(*argv)
		dst = cleanname(*argv++);
	if(!src || !dst)
		sysfatal(usage);
	if(!mflag && Rflag)
		sysfatal(usage);
	if(file && *argv)
		sysfatal(usage);

	/* check src whether it is not empty */
	switch(check(src)){
	case 1:
		sysfatal("%s is non-existent", src);
	case 2:
		sysfatal( "%s is not a directory", src);
	case 3:
		sysfatal( "%s is empty", src);
	default:
		break;
	}

	/* check dst whether it exits */
	switch(check(dst)){
	case 1: /* OK */
		if(mflag)
			sysfatal("%s absent. remove merge option -m",dst);
		break;
	case 2:
		sysfatal("%s is not a directory", dst);
	default: /*already exist */
		if(mflag)
			break;
		sysfatal("%s already exist. merge option -m is required", dst);
	}

	/* 	confirm that  dst is not in the subtree of src.
	*	if it is, we must abort processing    */
	if(strlen(dst) >= strlen(src) && strncmp(dst,src,strlen(src)) == 0){
		/* then be careful */
		if(dst[strlen(src)] == '/'){
			print("%s is a subtree of %s\n", dst, src);
			exits("error");
		}
	}

	if(xfile){
		int fd;
		Biobuf in;
		char buf[MAXPATH];
		char *line;
		char *path;
		int n;
		fd = open(xfile, OREAD);
		if(fd < 0)
			sysfatal("%s not open:%r", xfile);
		Binit(&in, fd, OREAD);
		while(line = Brdline(&in,'\n')){/* assign = */
			n = Blinelen(&in);
			strncpy(buf, line, n);
			buf[n] = 0;
			path = strtrim(buf);
			if(path[0] == '#')
				continue;
			path = cleanname(path);
			appendList(&xlist, path);
			//print("xfile... %s\n", path);
		}
		Bterm(&in); // note: Bterm close the file

		if(0) { // DEBUG
			List *p;
			p = xlist;
			while(p){
				print("%s\n", p->data);
				p = p->next;
			}
		}
	}

	if(!mflag) /* create the dir */
		mkdir(dst);

	/*	and process pathlist	*/
	if(file){
		Biobuf in;
		char *line;
		char buf[MAXPATH];
		int fd;
		int n;

		fd = open(file, OREAD);
		if(fd<0)
			sysfatal("%s not open\n", file);
		Binit(&in, fd, OREAD);
		while(line = Brdline(&in,'\n')){/* assign = */
			n = Blinelen(&in);
			strncpy(buf, line, n);
			buf[n] = 0;
			path = strtrim(buf);
			if(path[0] == '#')
				continue;
			path = cleanname(path);
			if(vflag) print("looking... %s\n", path);
			cpdir(path);
		}
		Bterm(&in); // note: Bterm close the file
	}
	else if(*argv)
		while(*argv){
			path = cleanname(*argv++);
			if(vflag) print("looking... %s\n", path);
			cpdir(path);
		}
	else
		cpdir("/");
}

void
cpdir(char *path)
{
	int sfd,dfd;
	int	nsrcdir;
	int	ndstdir;
	Dir *srcdirbuf, *sd, *sd0, *sdend, *ddend;
	Dir *dstdirbuf, *dd, ndd;
	char *newpath;
	char srcpath[MAXPATH];
	char dstpath[MAXPATH];

	if(path[0] == '/') path++;

	if(strcmp(src,"/") == 0)
		snprint(srcpath,sizeof(srcpath),"/%s", path);
	else
		snprint(srcpath,sizeof(srcpath),"%s/%s", src, path);

	//if(vflag) print("checking %s\n", path);
	if(xmatch(path)){
		if(vflag) print("skipping: %s\n", srcpath);
		return;
	}

	sfd = open(srcpath,OREAD);
	if(sfd == -1){
		fprint(2,"# open error: %s %r\n", srcpath);
		return;
	}
	sd = dirfstat(sfd);
	if(sd == nil){
		fprint(2,"# %s can't dirstat\n", srcpath);
		close(sfd);
		return;
	}
	if((sd->mode & DMDIR) == 0){ /* file */
		snprint(dstpath,sizeof(dstpath),"%s/%s", dst, path);
		dfd = open(dstpath,OREAD);
		if(dfd < 0){
			free(sd);
			close(sfd);
			cpfile(path);
			return;
		}
		dd = dirfstat(dfd);
		if(dd == nil){
			fprint(2,"# %s can't dirstat\n", dstpath);
			free(sd);
			close(sfd);
			close(dfd);
			return;
		}
		if(sd->mtime > dd->mtime || dd->mode & DMDIR){
			free(sd);
			close(sfd);
			close(dfd);
			cpfile(path);
			return;
		}
		if(sd->mtime == dd->mtime){
			nulldir(&ndd);
			if(copydirinfo(sd,&ndd) > 0 && dirfwstat(dfd,&ndd) < 0)
					fprint(2,"# can't wstat %s: %r\n", dstpath);
		}
		free(sd);
		free(dd);
		close(sfd);
		close(dfd);
		return;
	}
	nsrcdir = dirfinfo(&srcdirbuf, sfd);
	close(sfd);
	if(nsrcdir < 0){
		/* something wrong. skip this copy */
		fprint(2,"# %s unreadable\n", srcpath);
		free(sd);
		return;
	}

	snprint(dstpath,sizeof(dstpath),"%s/%s", dst, path);
	dfd = open(dstpath,OREAD);
	//print("open: %s %d\n", dstpath,dfd);
	if(dfd < 0){// try to create the dir
		if(mkdir(dstpath) < 0){
			fprint(2,"# unreadable to create %s: %r\n", dstpath);
			free(sd);
			return;
		}
		dfd = open(dstpath, OREAD);
		if(dfd < 0){
			fprint(2,"# created but not open %s: %r\n", dstpath);
			free(sd);
			return;
		}
	}
	sd0 = sd;
	/* end of flag f */

	ndstdir = dirfinfo(&dstdirbuf, dfd);
	if(ndstdir < 0){
		/* skip this copy */
		fprint(2, "# %s unreadable\n", dstpath);
		return;
	}
	sdend = srcdirbuf + nsrcdir;
	ddend = dstdirbuf + ndstdir;
	/* copy them */
	for(sd = srcdirbuf, dd = dstdirbuf; sd < sdend && dd < ddend; ){
		debug("comparing %s %s\n",sd->name, dd->name);
		if(strcmp(sd->name, dd->name) < 0){
			/* create destination */
			newpath = mkpath2(path, sd->name);
			if(sd->mode & DMDIR){
				/* create the directory and go on */
				char *p;
				p = mkpath2(dst, newpath);
				mkdir(p);
				free(p);
			}
			cpdir(newpath);
			free(newpath);
			sd++;
			continue;
		}
		else if(strcmp(sd->name, dd->name) > 0){
			if(Rflag){
				/* remove destination */
				char *p;
				p = mkpath3(dst, path, dd->name);
				if(vflag) print("removing: %s\n", p);
				if(dd->mode & DMDIR)
					rmdir(p);
				else
					rmfile(p);
				free(p);
			}
			dd++;
			continue;
		}
		else /* both names exist */
		if((sd->mode & DMDIR) && (dd->mode & DMDIR)){
			/* 	both are directories, then we must go farther.
			*	time info is no help! */
			newpath = mkpath2(path, sd->name);
			cpdir(newpath);
			free(newpath);
		}
		else /* one or both of two is a file */
		if(sd->mtime >= dd->mtime) {
			char *target;
			newpath = mkpath2(path, sd->name);
			target = mkpath3(dst, path, sd->name);
			if(sd->mode & DMDIR){ /* the source is a directory */
				int err = 0;
				int fd;
				if(rmfile(target) < 0)
					err = 1;
				fd = create(target, OREAD, sd->mode);
				if(fd < 0){
					fprint(2,"# unable to create %s: %r", target);
					err = 1;
				}
				else
					close(fd);
				if(!err) cpdir(newpath);
			}
			else{ /* the source is a file. */
				if(dd->mode & DMDIR){ /* target is a dir. */
					rmdir(target);
					dd->mtime = 0;
				}
				if(sd->mtime > dd->mtime)
					cpfile(newpath);
				else{ /* sd->mtime == dd->mtime */
					if(copydirinfo(sd,dd) > 0){
						if(dirwstat(target, dd) < 0)
							fprint(2,"# %s unable to wstat\n", target);
					}
				}
			}
			free(newpath);
			free(target);
		}
		else
			if(sd->mtime < dd->mtime){
				newpath = mkpath3(dst, path,dd->name);
				fprint(2,"# %s untouched\n", newpath);
				free(newpath);
			}
		sd++;
		dd++;
	}

	if(sd < sdend){
		char *p;
		for(; sd < sdend; sd++){
			newpath = mkpath2(path, sd->name);
			p = mkpath3(dst, path, sd->name);
			if(sd->mode & DMDIR){
				/* create the directory and go on */
				mkdir(p);
				cpdir(newpath);
			}
			else
				cpfile(newpath);
			free(newpath);
			free(p);
		}
	}
	else if(Rflag && dd < ddend){
		char *p;
		for(; dd < ddend; dd++){
			p = mkpath3(dst, path, dd->name);
			if(vflag) print("removing: %s\n", p);
			if(dd->mode & DMDIR)
				rmdir(p);
			else
				rmfile(p);
			free(p);
		}
	}

	/* free dirbuf */
	free(srcdirbuf);
	free(dstdirbuf);
	nulldir(&ndd);
	if(copydirinfo(sd0,&ndd) > 0 && dirfwstat(dfd, &ndd) < 0)
		fprint(2,"# could not wstat %s: %r\n", dstpath);
	close(dfd);
	free(sd0);
}

int
check(char *path)
{
	int fd;
	Dir *db, *d;
	fd = open(path, OREAD);
	if(fd == -1)
		return 1; // non existent
	d = dirfstat(fd);
	if((d->mode & DMDIR) == 0){
		close(fd);
		free(d);
		return 2; // not a directory
	}
	free(d);
	if(dirread(fd, &db) == 0){
		close(fd);
		free(db);
		return 3; // empty
	}
	close(fd);
	free(db);
	return 0;
}



/*
*	dirinfo
*
*	return value
*	case -1: not exit
*	case 0: empty
*	others: the numbers of files/directory
*
*	note:
*		free(dirbuf) if return value is not -1
*
*	ref: /sys/src/cmd/ls.c
*/
int
dirinfo(Dir **dirbufp, char *path)
{
	int fd;
	Dir *db;
	int n;

	fd = open(path, OREAD);
	if(fd == -1)
		return -1;
	n = dirreadall(fd,&db);
	close(fd);
	if(n == -1)
		return -1;

	/* sort by name */
	qsort(db, n, sizeof(Dir), (int (*)(void*, void*))compar);

	*dirbufp = db;
	return n;
}

int
dirfinfo(Dir **dirbufp, int fd)
{
	Dir *db;
	int n;

	n = dirreadall(fd,&db);
	if(n == -1)
		return -1;

	/* sort by name */
	qsort(db, n, sizeof(Dir), (int (*)(void*, void*))compar);

	*dirbufp = db;
	return n;
}


int
compar(Dir *a, Dir *b)
{
	return strcmp(a->name, b->name);
}

int
mkdir(char *dstpath)
/*	dstpath: absolute path to the destination	*/
{
	int fd;
	char *p, *q;
	if(vflag) print("creating: %s\n", dstpath);
	fd = create(dstpath, OREAD, DMDIR | 0775);
	if(fd < 0){
		p = dstpath + strlen(dst);
		while((q = strchr(p, '/'))){ // assign =
			*q = 0;
			switch(check(dstpath)){
			case 2:
				rmfile(dstpath); // continue to next case
			case 1:
				fd = create(dstpath, OREAD, DMDIR | 0775);
				if(fd < 0){
					fprint(2,"# can't create %s: %r\n", dstpath);
					return -1;
				}
				close(fd);
				break;
			default:
				break;
			}
			*q = '/';
			p = q + 1;
		}
		fd = create(dstpath, OREAD, DMDIR | 0775);
		if(fd < 0){
			fprint(2,"# can't create %s: %r\n", dstpath);
			return -1;
		}
	}
	close(fd);
	return 0;
}

int
copyfile(int sfd, int dfd)
{
	char buf[4096];
	int n;
	for(;;){
		n = read(sfd, buf, sizeof buf);
		if(n <= 0) break;
		if(write(dfd, buf, n) != n)
			return -1;
	}
	return 0;
}

int
copydirinfo(Dir *sd, Dir *dd)
{
	int f = 0;
	/* 	&& ++f may be tricky
	*	case dirflag
	*		1: directory
	*		0: file		*/
	if(dd->mode != sd->mode && ++f)
		dd->mode = sd->mode;
	if(dd->mtime != sd->mtime && ++f)
		dd->mtime = sd->mtime;
	if(gflag && strcmp(dd->gid,sd->gid)!=0 && ++f)
		dd->gid = sd->gid;
	if(uflag && strcmp(dd->uid,sd->uid)!=0 && ++f)
		dd->uid = sd->uid;
	return f;
}

/*	src/path should be a file	*/
void
cpfile(char *path)
{
	char srcpath[MAXPATH];
	char dstpath[MAXPATH];
	Dir *sd, *dd, ndd;
	int sfd, dfd;
	if(path[0] == '/') path++;

	if(strcmp(src,"/") == 0)
		snprint(srcpath,sizeof(srcpath),"/%s", path);
	else
		snprint(srcpath,sizeof(srcpath),"%s/%s", src, path);

	//if(vflag) print("checking %s\n", path);
	if(xmatch(path)){
		if(vflag) print("skipping: %s\n", srcpath);
		return;
	}

	snprint(dstpath,sizeof(dstpath),"%s/%s", dst, path);
	sfd = open(srcpath, OREAD);
	if(sfd < 0){
		fprint(2,"# open error: %s %r\n", srcpath);
		return;
	}
	sd = dirfstat(sfd);
	if(!sd){ // something wrong
		fprint(2,"# cannot read stat: %s\n", srcpath);
		close(sfd);
		return;
	}
	if(sd->mode & DMDIR){ //illegal usage
		fprint(2,"# bug of this program\n");
		close(sfd);
		free(sd);
		return;
	}
	dfd = open(dstpath, OWRITE|OTRUNC);
	if(dfd < 0){
		dfd = create(dstpath, OWRITE | OTRUNC, 0777);
		if(dfd < 0){
			fprint(2,"# unable to create %s: %r\n", dstpath);
			close(sfd);
			free(sd);
			return;
		}
	}
	dd = dirfstat(dfd);
	if(!dd){ // something wrong
		fprint(2,"# cannot read stat: %s\n", dstpath);
		close(sfd);
		free(sd);
		close(dfd);
		return;
	}
	if(dd->mode & DMAPPEND){
		nulldir(&ndd);
		ndd.mode = dd->mode & ~DMAPPEND;
		if(dirfwstat(dfd,&ndd) < 0){
			fprint(2,"# cannot chmod -a : %s\n", dstpath);
			close(sfd);
			free(sd);
			close(dfd);
			free(dd);
			return;
		}
		close(dfd);
		free(dd);
		dfd = create(dstpath, OWRITE | OTRUNC, 0777);
		if(dfd < 0){
			fprint(2,"# unable to create %s: %r\n", dstpath);
			close(sfd);
			free(sd);
			return;
		}
		dd = dirfstat(dfd);
		if(!dd){ // something wrong
			fprint(2,"# cannot read stat: %s\n", dstpath);
			close(sfd);
			free(sd);
			close(dfd);
			return;
		}
	}
	if(dd->mode & DMDIR){
		remove(dstpath);
		close(dfd);
		dfd = create(dstpath, OWRITE | OTRUNC, 0777);
		if(dfd < 0){
			fprint(2, "# unable to create %s: %r\n", dstpath);
			close(sfd);
			free(sd);
			free(dd);
			return;
		}
	}
	if(vflag) print("copying: %s\n", srcpath);
	if(copyfile(sfd,dfd) < 0){
		fprint(2, "# write error: %s\n", dstpath);
		close(sfd);
		free(sd);
		close(dfd);
		free(dd);
		return;
	}

	nulldir(&ndd);
	if(copydirinfo(sd,&ndd) > 0 && dirfwstat(dfd,&ndd)<0)
		fprint(2, "# can't wstat %s: %r\n", dstpath);
	free(sd);
	free(dd);
	close(sfd);
	close(dfd);
	return;
}

int
rmfile(char *path){
	int status;
	status = remove(path);
	if(status == -1)
		fprint(2, "# %s not removed", path);
	return status;
}

int
rmdir(char *path)
{
	char *name;
	int i, ndir;
	Dir *dirbuf;
	int status;

	ndir = dirinfo(&dirbuf, path);
	if(ndir < 0){
		fprint(2, "# cannot get dir info %s\n", path);
		return -1;
	}

	for(i=0; i<ndir; i++){
		name = mkpath2(path, dirbuf[i].name);
		if(dirbuf[i].mode & DMDIR)
			rmdir(name);
		else
			rmfile(name);
		free(name);
		
	}
	status = rmfile(path);
	free(dirbuf);
	return status;
}
